import time
import configobj
import logging
import warnings, functools
try: import simplejson as json
except ImportError: import json
from datetime import datetime

import sys
if sys.version_info < (3,):
    def deprecated(func):
        '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.'''

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn_explicit(
                "Call to deprecated function {}.".format(func.__name__),
                category=DeprecationWarning,
                filename=func.func_code.co_filename,
                lineno=func.func_code.co_firstlineno + 1
            )
            return func(*args, **kwargs)
        return new_func
else:
    def deprecated(func):
        '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.'''

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn_explicit(
                "Call to deprecated function {}.".format(func.__name__),
                category=DeprecationWarning,
                filename=func.__code__.co_filename,
                lineno=func.__code__.co_firstlineno + 1
            )
            return func(*args, **kwargs)
        return new_func

from .senderprotocol import SenderProtocol

class DataContainer(SenderProtocol):

    REQUEST = "sender data"
    _data_type = None
    _items_list = []

    def __init__(self, data_type  = None,
                       zbx_file   = '/etc/zabbix/zabbix_agentd.conf',
                       zbx_host   = None,
                       zbx_port   = None,
                       log_level  = None,
                       log_output = None,
                       dryrun     = False,
                       logger     = None):
        super(DataContainer,self).__init__()


        # Loads config from zabbix_agentd file
        # If no file, it uses the default _config as configuration
        self._load_zabbix_config(zbx_file)
        # Override default values with the ones provided
        if log_level:
            self._config['log_level'] = log_level
        if log_output != None:
            self._config['log_output'] = log_output
        self._config['dryrun'] = dryrun
        if zbx_host:
            self._config['server'] = zbx_host
        if zbx_port:
            self._config['port'] = zbx_port
        if data_type:
            self._config['data_type'] = data_type
        self._logger = logger

    @property
    def data_type(self):
        return self._config['data_type']

    @data_type.setter
    def data_type(self, value):
        if self._logger:
            self._logger.debug("Setting value %s as data_type" % value)
        if value in ['lld', 'items']:
          self._config['data_type'] = value
          # Clean _items_list & _result when changing _data_type
          # Incompatible format
          self._items_list = []
          self._result = []
        else:
            if self._logger:
                self._logger.error('data_type requires either "items" or "lld"')
            raise ValueError('data_type requires either "items" or "lld"')

    @property
    def log_level(self):
        return int(self._config['log_level'])

    @log_level.setter
    def log_level(self, value):
        if isinstance(value, int) and value >= 0 and value < 5:
            self._config['log_level'] = value
        else:
            if self._logger:
                self._logger.error('log_level parameter must be less than 5')
            raise ValueError('log_level parameter must be less than 5')

    @property
    def dryrun(self):
        return self._config['dryrun']

    @dryrun.setter
    def dryrun(self, value):
        if value in [True, False]:
            self._config['dryrun'] = value
        else:
            if self._logger:
                self._logger.error('dryrun parameter requires boolean')
            raise ValueError('dryrun parameter requires boolean')

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, value):
        if isinstance(value, logging.Logger):
            self._logger = value
        else:
            if self._logger:
                self._logger.error('logger requires a logging instance')
            raise ValueError('logger requires a logging instance')

    @property
    def log_output(self):
        return self._config['log_output']

    def _load_zabbix_config(self,config_file):
        # Load zabbix agent configuration as default values
        # Default values are set in self._config
        # - ServerActive (default: 127.0.0.1)
        # - LogFile (default: /tmp/zabbix_agentd.log)
        # - DebugLevel (default: 3, Allowed: 0-4)
        #               0 -> logging.NOTSET
        #               1 -> logging.CRITICAL
        #               2 -> logging.ERROR
        #               3 -> logging.WARNING
        #               4 -> logging.DEBUG
        # - Timeout (default: 3, Allowed: 1-30)
        tmp_config = configobj.ConfigObj(config_file, list_values=False)

        if 'ServerActive' in tmp_config:
            tmp_server = tmp_config['ServerActive'][0] \
                         if isinstance(tmp_config['ServerActive'], list) \
                         else list(tmp_config['ServerActive'])[0]
            self._config['server'], self._config['port'] = tmp_server.split(':') \
                         if ":" in tmp_server else (tmp_server, 10051)

        if 'LogFile' in tmp_config:
            self._config['log_output'] = tmp_config['LogFile']

        if 'DebugLevel' in tmp_config:
            self._config['log_level'] = int(tmp_config['DebugLevel'])

        if 'Timeout' in tmp_config:
            self._config['timeout'] = int(tmp_config['Timeout'])

    def add_item(self, host, key, value, clock=None):
        if clock is None:
            clock = self.clock
        if self._config['data_type'] == "items":
            item = { "host": host, "key": key,
                     "value": value, "clock": clock}
        elif self._config['data_type'] == "lld":
            item = { "host": host, "key": key, "clock": clock,
                     "value": json.dumps({"data": value}) }
        else:
            if self._logger:
                self._logger.error('Setup data_type before adding data')
            raise ValueError('Setup data_type before adding data')
        self._items_list.append(item)

    def add(self, data):
        for host in data:
            for key in data[host]:
                if not data[host][key] == []:
                    self.add_item( host, key, data[host][key])

    @deprecated
    def set_type(self, value):
        self.data_type = value



import re
import socket
import struct
import time
import sys
import warnings,functools
try: import simplejson as json
except ImportError: import json

if sys.version_info < (3,):
    def b(x):
        return x
else:
    import codecs
    def b(x):
        return codecs.utf_8_encode(x)[0]

if sys.version_info < (3,):
    def deprecated(func):
        '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.'''

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn_explicit(
                "Call to deprecated function {}.".format(func.__name__),
                category=DeprecationWarning,
                filename=func.func_code.co_filename,
                lineno=func.func_code.co_firstlineno + 1
            )
            return func(*args, **kwargs)
        return new_func
else:
    def deprecated(func):
        '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.'''

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.warn_explicit(
                "Call to deprecated function {}.".format(func.__name__),
                category=DeprecationWarning,
                filename=func.__code__.co_filename,
                lineno=func.__code__.co_firstlineno + 1
            )
            return func(*args, **kwargs)
        return new_func

ZBX_HDR = "ZBXD\1"
ZBX_HDR_SIZE = 13
# For both 2.0 & >2.2 Zabbix version
# ? 1.8: Processed 0 Failed 1 Total 1 Seconds spent 0.000057
# 2.0: Processed 0 Failed 1 Total 1 Seconds spent 0.000057
# 2.2: processed: 50; failed: 1000; total: 1050; seconds spent: 0.09957
# 2.4: processed: 50; failed: 1000; total: 1050; seconds spent: 0.09957
ZBX_RESP_REGEX = r'[Pp]rocessed:? (\d+);? [Ff]ailed:? (\d+);? [Tt]otal:? (\d+);? [Ss]econds spent:? (\d+\.\d+)'
ZBX_DBG_SEND_RESULT = "Send result [%s-%s-%s] for %s"
ZBX_DBG_SEND_ITEM   = "[%s %s %s]"
ZBX_SEND_ITEM   = "[%d items]"

class SenderProtocol(object):
    def __init__(self):
        self._config = {
            'server': '127.0.0.1',
            'port': 10051,
            'log_output': '/tmp/zabbix_agentd.log',
            'log_level': 3,
            'timeout': 3,
            'dryrun': False,
            'data_type': None }
        self._items_list = []
        self.data = None
    
    @property
    def zbx_host(self):
        return self._config['server']

    @zbx_host.setter
    def zbx_host(self, value):
        self._config['server'] = value

    @property
    def zbx_port(self):
        return self._config['port']

    @zbx_port.setter
    def zbx_port(self, value):
        if isinstance(value, int) and \
           value > 0 and value < 65535:
            self._config['port'] = value
        else:
            if self._logger:
                self._logger.error('zbx_port requires a valid TCP port number')
            raise ValueError('zbx_port requires a valid TCP port number')

    @property
    def items_list(self):
        return self._items_list

    @property
    def result(self):
        return self._result

    @property
    def clock(self):
        return int(time.time())

    def _send_to_zabbix(self, item):
        # Return 0 if dryrun mode enabled
        if self._config['dryrun']:
            return 0
        # Format data to be sent
        if type(item) is dict:
            item = [ item ]
        self.data = json.dumps({ "data": item,
                                 "request": self.REQUEST,
                                 "clock": self.clock })
        # Set socket options & open connection
        socket.setdefaulttimeout(self._config['timeout'])
        try:
            self.zbx_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.zbx_sock.connect((self._config['server'], int(self._config['port'])))
        except Exception as e:
            # Maybe we could consider storing missed sent data for later retry
            self.data = None
            self._items_list = []
            self.zbx_sock.close()
            if self._logger:
                self._logger.error(
                    "Unable to connect to server %s on port %d" % \
                    (self._config['server'], self._config['port'])
                )
            raise
        # Build Zabbix Sender payload
        data_length = len(self.data)
        data_header = struct.pack('<Q', data_length)
        packet = b(ZBX_HDR) + data_header + b(self.data)
        # Send payload to Zabbix Server and check response header
        try:
            self.zbx_sock.sendall(packet)
        except:
            if self._logger:
                self._logger.error('Error while sending data to Zabbix server')
            raise

        try:
            # Check the 5 first bytes from answer to ensure it's well formatted
            zbx_srv_resp_hdr = self.zbx_sock.recv(5)
            assert(zbx_srv_resp_hdr == b(ZBX_HDR))
        except Exception as e:
            if self._logger:
                self._logger.error(
                    'Error while receiving response from Zabbix server [%s]' % e
                )
            raise
        # Get the 8 next bytes and unpack to get response's payload length
        zbx_srv_resp_data_hdr = self.zbx_sock.recv(8)
        zbx_srv_resp_body_len = struct.unpack('<Q', zbx_srv_resp_data_hdr)[0]
        # Get response payload from Zabbix Server
        zbx_srv_resp_body = self.zbx_sock.recv(zbx_srv_resp_body_len)
        self.data = None
        self._items_list = []
        self.zbx_sock.close()
        if sys.version_info[0] == 3:
            zbx_srv_resp_body = zbx_srv_resp_body.decode()
        return json.loads(zbx_srv_resp_body)

    def send(self, container = None):
        if container != None and self.logger:
            # Using container argument is deprecated
            self.logger.warning(
                'Deprecated call of send() function with container argument'
            )
        zbx_answer = 0
        self._result = []
        if self._config['log_level'] >= 4:
            # Per item sent if debug mode enabled
            for item in self._items_list:
                output =  ZBX_DBG_SEND_ITEM % (
                    item["host"],
                    item["key"],
                    item["value"]
                )
                zbx_answer = self._send_to_zabbix(item)
                result = self._handle_response(zbx_answer, output)
                if self.logger:
                    self._logger.debug(
                        ZBX_DBG_SEND_RESULT % (
                            result[0],
                            result[1],
                            result[2],
                            output
                        )
                    )
                self._result.append(result)
        else:
            # All items at once if no debug
            output = ZBX_SEND_ITEM % (
                len(self._items_list)
            )
            zbx_answer = self._send_to_zabbix(self._items_list)
            result = self._handle_response(zbx_answer, output)
            if self.logger:
                self._logger.info(
                    ZBX_DBG_SEND_RESULT % (
                        result[0],
                        result[1],
                        result[2],
                        output
                    )
                )
            self._result.append(result)
        self.data = None
        self._items_list = []
        if not self._config['dryrun']:
            self.zbx_sock.close()
        self._config['data_type'] = None

    def _handle_response(self, zbx_answer, output):
        if zbx_answer and self.logger:
            self.logger.debug("Got [%s] as response from Zabbix server" % zbx_answer)
        nb_item = len(self._items_list)
        if self._config['log_level'] >= 4:
            nb_item = 1
        if zbx_answer:
            if zbx_answer.get('response') == 'success':
                result = re.findall( ZBX_RESP_REGEX, zbx_answer.get('info'))
                result = result[0]
        else:
            result = ['d', 'd', str(nb_item)]
        return result

    @deprecated
    def set_debug(self, value):
        if value:
            self._config['log_level'] = 4
        else:
            self._config['log_level'] = 3

    @deprecated
    def set_verbosity(self, value):
        pass

    @deprecated
    def set_dryrun(self, value):
        if value == None:
            value = False
        self._config['dryrun'] = value

    @deprecated
    def set_host(self, value):
        self._config['server'] = value

    @deprecated
    def set_port(self, value):
        self._config['port'] = value

    @property
    @deprecated
    def debug(self):
        if self._config['log_level'] >= 4:
            return True
        else:
            return False

    @debug.setter
    @deprecated
    def debug(self, value):
        if value in [True, False]:
            if value is True:
                self._config['log_level'] = 4
            else:
                self._config['log_level'] = 3
        else:
            raise ValueError('debug parameter requires boolean')



#!/usr/bin/env python

import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    user_options = []
    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import pytest
        options = "--cov protobix --cov-report term-missing"
        try: import coverage
        except ImportError: options = ""
        errno = pytest.main(options)
        raise SystemExit(errno)

setup(
    name = 'protobix',
    packages = ['protobix'],
    version = '0.1.1rc5',
    install_requires = [
        'configobj',
        'simplejson',
        'traceback2'
    ],

    description = 'Implementation of Zabbix Sender protocol',
    long_description = ( 'This module implements Zabbix Sender Protocol.\n'
                         'It allows to build list of items and send '
                         'them as trapper.\n'
                         'It currently supports items update as well as '
                         'Low Level Discovery.' ),    
    author = 'Jean Baptiste Favre',
    author_email = 'jean-baptiste.favre@blablacar.com',
    license = 'GPL-3+',
    url='https://github.com/jbfavre/python-protobix/',
    download_url = 'https://github.com/jbfavre/python-protobix/archive/0.1.1rc5.tar.gz',
    keywords = ['monitoring','zabbix','trappers'],
    classifiers = [],
    cmdclass={'test': PyTest}
   )



import logging
import optparse
import socket
import sys
import traceback
from logging import handlers

from .datacontainer import DataContainer

class SampleProbe(object):

    __version__ = '0.0.9'
    # Mapping between zabbix-agent Debug option & logging level
    LOG_LEVEL = [
        logging.NOTSET,
        logging.CRITICAL,
        logging.ERROR,
        logging.INFO,
        logging.DEBUG
    ]
    logger = None
    probe_config = None

    def _parse_args(self):
        # Parse the script arguments
        # Common part
        parser = optparse.OptionParser()

        parser.add_option('-c', '--config', default = None,
                          help='Probe config file location. '
                               'Can be either absolute or relative path')
        parser.add_option('-d', '--dry', action='store_true', default = False,
                          help='Performs CDH API calls but do not send '
                               'anything to the Zabbix server. Can be used '
                               'for both Update & Discovery mode')
        parser.add_option('-D', '--debug', action='store_true', default = False,
                          help='Enable debug mode. This will prevent bulk send '
                               'operations and force sending items one after the '
                               'other, displaying result for each one')

        zabbix_options = optparse.OptionGroup(parser, 'Zabbix configuration')
        zabbix_options.add_option('-z', '--zabbix-server', default='127.0.0.1',
                                  help='The hostname of Zabbix server or '
                                       'proxy, default is 127.0.0.1.')
        zabbix_options.add_option('-p', '--zabbix-port', default=10051,
                                  help='The port on which the Zabbix server or '
                                       'proxy is running, default is 10051.')
        zabbix_options.add_option('--update-items', action='store_const',
                                  dest='mode', const='update_items',
                                  help='Get & send items to Zabbix. This is the default '
                                       'behaviour even if option is not specified')
        zabbix_options.add_option('--discovery', action='store_const',
                                  dest='mode', const='discovery',
                                  help='If specified, will perform Zabbix Low Level '
                                       'Discovery on Hadoop Cloudera Manager API. '
                                       'Default is to get & send items')
        parser.add_option_group(zabbix_options)
        parser.set_defaults(mode='update_items')

        return parser

    def _setup_logging(self, zbx_container):
        logger = logging.getLogger(self.__class__.__name__)
        common_log_format = '[%(name)s:%(levelname)s] %(message)s'
        # Enable default console logging
        # Only if we have a tty (zabbix-agent redirects stdout)
        if sys.stdout.isatty():
            consoleHandler = logging.StreamHandler()
            consoleFormatter = logging.Formatter(
                fmt = common_log_format,
                datefmt = '%Y%m%d:%H%M%S'
            )
            consoleHandler.setFormatter(consoleFormatter)
            logger.addHandler(consoleHandler)
        # zabbix_agent uses syslog if LogFile empty, specified file otherwise
        if zbx_container.log_output:
            try:
                # TODO use Datacontainer.log_output property
                fileHandler = logging.FileHandler('/tmp/zabbix_agentd.log')
                # Use same date format as Zabbix: when logging into
                # zabbix_agentd log file, it's easier to read & parse
                logFormatter = logging.Formatter(
                    fmt = '%(process)d:%(asctime)s.%(msecs)03d ' + common_log_format,
                    datefmt = '%Y%m%d:%H%M%S'
                )
                fileHandler.setFormatter(logFormatter)
                logger.addHandler(fileHandler)
            except IOError:
                pass
        else:
            syslogHandler = logging.handlers.SysLogHandler(
                address = '/dev/log',
                facility = logging.handlers.SysLogHandler.LOG_DAEMON
            )
            # Use same date format as Zabbix does: when logging into
            # zabbix_agentd log file, it's easier to read & parse
            logFormatter = logging.Formatter(
                fmt = '%(process)d:%(asctime)s.%(msecs)03d ' + common_log_format,
                datefmt = '%Y%m%d:%H%M%S'
            )
            syslogHandler.setFormatter(logFormatter)
            logger.addHandler(syslogHandler)
        logger.setLevel(
            self.LOG_LEVEL[zbx_container.log_level]
        )
        return logger

    def _init_container(self):
        zbx_container = DataContainer(
            data_type = 'items',
            zbx_host  = self.options.zabbix_server,
            zbx_port  = int(self.options.zabbix_port),
            log_level = 4 if self.options.debug else 3,
            dryrun    = self.options.dry,
            logger    = self.logger
        )
        return zbx_container

    def run(self):
        (self.options, args) = self._parse_args()
        self.hostname = socket.getfqdn()

        # Datacontainer init (needs logger)
        zbx_container = self._init_container()
        # logger init
        self.logger = zbx_container.logger = self._setup_logging(zbx_container)

        # Step 1: read probe configuration
        try:
            self._init_probe()
        except Exception as e:
            self.logger.error('Step 1 - Read probe configuration failed')
            self.logger.debug(traceback.format_exc())
            return 1

        # Step 2: get data
        try:
            data = {}
            if self.options.mode == "update_items":
                zbx_container.data_type = 'items'
                data = self._get_metrics()
            elif self.options.mode == "discovery":
                zbx_container.data_type = 'lld'
                data = self._get_discovery()
        except Exception as e:
            self.logger.error(
                'Step 2 - Get Data failed [%s]' % str(e)
            )
            self.logger.debug(traceback.format_exc())
            return 2

        # Step 3: format & load data into container
        try:
            zbx_container.add(data)
        except Exception as e:
            self.logger.error(
                'Step 3 - Format & add Data failed [%s]' % e.strerror
            )
            self.logger.debug(traceback.format_exc())
            return 3

        # Step 4: send container data to Zabbix server
        try:
            zbx_container.send()
        except socket.error as e:
            self.logger.error(
                'Step 4 - Sent to Zabbix Server failed [%s]' % str(e)
            )
            self.logger.debug(traceback.format_exc())
            return 4
        except Exception as e:
            self.logger.error(
                'Step 4 - Unknown error [%s]' % e.strerror
            )
            self.logger.debug(traceback.format_exc())
            return 4
        # Everything went fine. Let's return 0 and exit
        return 0
